/*
 *  ClipPlayer.java
 *
 *  This file is part of the Java Sound Examples.
 */

/*
 *  Copyright (c) 1999 - 2001 by Matthias Pfisterer <Matthias.Pfisterer@web.de>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Library General Public License as published
 *   by the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU Library General Public License for more details.
 *
 *   You should have received a copy of the GNU Library General Public
 *   License along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */


import  java.io.File;
import  java.io.IOException;
import  java.io.ByteArrayInputStream;

import  javax.sound.sampled.AudioFileFormat;
import  javax.sound.sampled.AudioFileFormat.Type;
import  javax.sound.sampled.AudioFormat;
import  javax.sound.sampled.AudioInputStream;
import  javax.sound.sampled.AudioSystem;
import  javax.sound.sampled.Clip;
import  javax.sound.sampled.DataLine;
import  javax.sound.sampled.LineEvent;
import  javax.sound.sampled.LineListener;
import  javax.sound.sampled.LineUnavailableException;


public class ClipPlayer
    implements  LineListener
{
    private Clip        m_clip;
    private int         m_nLoopCount;
    private byte[]      m_buffer;
    private AudioFormat m_format;
    private int         m_frameSize;


    /*
     *  The clip will be played nLoopCount + 1 times.
     */
    public ClipPlayer(File clipFile, int nLoopCount)
    {
        m_buffer = null;
        m_nLoopCount = nLoopCount;
        
        AudioInputStream    audioInputStream = null;
        try
        {
            audioInputStream = AudioSystem.getAudioInputStream(clipFile);
            // let's find out what's up here...
            pl("the # of available bytes w/o blocking is " + audioInputStream.available());
            pl("the frame length of the whole stream is " + audioInputStream.getFrameLength());
            pl("the markSupported status of this a. i. stream is " + audioInputStream.markSupported());
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
        if (audioInputStream != null)
        {
            AudioFormat format = audioInputStream.getFormat();
            m_format = format;
            // for fun. let's print what's going on in this format...
            System.out.println("The format is  " + format);
            pl("the number of bytes per frame is " + format.getFrameSize());
            
            // let's copy all of the audio input stream's data to a buffer!
            int frameLength = (int)(audioInputStream.getFrameLength());
            int frameSize = format.getFrameSize();
            m_frameSize = frameSize;
            int bufferSize = frameLength * frameSize;
            pl(" the buffer size is " + bufferSize);
            m_buffer = new byte[bufferSize];
            
            try
            {
                int nReadBytes = audioInputStream.read(m_buffer,0,m_buffer.length);
                pl("number of read bytes was " + nReadBytes);
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
            
            // now, the buffer is full o' data -- let's reverse it 
            // NOTE this has to be done one __frame__ at a time...
            
            byte[] revBuffer = new byte[m_buffer.length];
            int rf=0; // reversed buffer frame index
            
            // notice that direct backwardsing doesn't work!!
            // you don't want to reverse the frames!
            for (int of = frameLength-1 ; of>=0 ; --of) // frames in the original buffer
            {
                revBuffer[2*rf+0] = m_buffer[2*of+0];
                revBuffer[2*rf+1] = m_buffer[2*of+1];
                ++rf;
            }
            
            // switch buffers
            m_buffer = revBuffer;
            
            DataLine.Info   info = new DataLine.Info(Clip.class, format);
            try
            {
                m_clip = (Clip) AudioSystem.getLine(info);
                m_clip.addLineListener(this);
                
                // this next line of code opens the clip for playing the data from the
                // audio input stream BUT we want to use the alternative open
                // provided for clips that allows us to write all the data at once!
                // m_clip.open(audioInputStream);
                
                // so, let's read data from the audio input stream and write it
                // to a byte buffer
                m_clip.open(m_format, m_buffer, 0, m_buffer.length);
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
            
            //transferClip();
            m_clip.loop(m_nLoopCount);
            
            // let's keep going!
            // the clip still exists, right?
            // first we create an input stream from the buffer...
            ByteArrayInputStream bais = new ByteArrayInputStream(m_buffer);
            AudioInputStream ais = new AudioInputStream(bais,m_format,(long)(m_buffer.length/m_frameSize));
            File fileOut = new File("PierceOut.wav");
            try
            {
                AudioSystem.write(ais,AudioFileFormat.Type.WAVE,fileOut);
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
            pl("Write to PierceOut.wav");
            
            /* freeware to convert mp3 to wav and back:
             *   http://www.hitsquad.com/smm/programs/Mobsoft_Mp3toWav/
             */
            
            /* lots of code !! at http://www.musicdsp.org/files/(pink.txt)
             *
             * here's pink noise:
             *
             Filter to make pink noise from white  (updated March 2000)
------------------------------------

This is an approximation to a -10dB/decade filter using a weighted sum
of first order filters. It is accurate to within +/-0.05dB above 9.2Hz 
(44100Hz sampling rate). Unity gain is at Nyquist, but can be adjusted
by scaling the numbers at the end of each line.

If 'white' consists of uniform random numbers, such as those generated
by the rand() function, 'pink' will have an almost gaussian level 
distribution.


  b0 = 0.99886 * b0 + white * 0.0555179;
  b1 = 0.99332 * b1 + white * 0.0750759;
  b2 = 0.96900 * b2 + white * 0.1538520;
  b3 = 0.86650 * b3 + white * 0.3104856;
  b4 = 0.55000 * b4 + white * 0.5329522;
  b5 = -0.7616 * b5 - white * 0.0168980;
  pink = b0 + b1 + b2 + b3 + b4 + b5 + b6 + white * 0.5362;
  b6 = white * 0.115926;


An 'economy' version with accuracy of +/-0.5dB is also available.

  b0 = 0.99765 * b0 + white * 0.0990460;
  b1 = 0.96300 * b1 + white * 0.2965164;
  b2 = 0.57000 * b2 + white * 1.0526913;
  pink = b0 + b1 + b2 + white * 0.1848;
  
            *
            * here's a lowpass and highpass filter: http://www.musicdsp.org/showArchiveComment.php?ArchiveID=38
            *
 LP and HP filter

Type : biquad, tweaked butterworth
References : Posted by Patrice Tarrabia
Code : 
r = rez amount, from sqrt(2) to ~ 0.1
f = cutoff frequency
(from ~0 Hz to SampleRate/2 - though many
synths seem to filter only up to SampleRate/4)

The filter algo:
out(n) = a1 * in + a2 * in(n-1) + a3 * in(n-2) - b1*out(n-1) - b2*out(n-2)

Lowpass:
c = 1.0 / tan(pi * f / sample_rate);

a1 = 1.0 / ( 1.0 + r * c + c * c);
a2 = 2* a1;
a3 = a1;
b1 = 2.0 * ( 1.0 - c*c) * a1;
b2 = ( 1.0 - r * c + c * c) * a1;

Hipass:
c = tan(pi * f / sample_rate);

a1 = 1.0 / ( 1.0 + r * c + c * c);
a2 = -2*a1;
a3 = a1;
b1 = 2.0 * ( c*c - 1.0) * a1;
b2 = ( 1.0 - r * c + c * c) * a1;

Added on : 01/01/70 by andy_rossol[AT]hotmail.com
Comment : 
Ok, the filter works, but how to use the resonance parameter (r)? 
The range from sqrt(2)-lowest to 0.1 (highest res.) is Ok 
for a LP with Cutoff > 3 or 4 KHz, but for lower cutoff frequencies 
and higher res you will get values much greater than 1! (And this means clipping like hell)

So, has anybody calculated better parameters (for r, b1, b2)?

            *
            *
            */

        }
        else
        {
            System.out.println("ClipPlayer.<init>(): can't get data from file " + clipFile.getName());
        }
    }
    
    public void transferClip()
    {
         m_clip.loop(m_nLoopCount);   
    }
    
    public void pl(Object o)
    {
        System.out.println("" + o);
    }
    
    public void p(Object o)
    {
        System.out.print("" + o);    
    } 



    public void update(LineEvent event)
    {
        if (event.getType().equals(LineEvent.Type.STOP))
        {
            m_clip.close();
        }
        else if (event.getType().equals(LineEvent.Type.CLOSE))
        {
            /*
             *  There is a bug in the jdk1.3.0.
             *  It prevents correct termination of the VM.
             *  So we have to exit ourselves.
             */
            System.exit(0);
        }

    }



    //public static void main(String[] args)
    public static void main(int i,String[] args)
    {
        //System.out.println("args.length is " + args.length);
        if (i == 0)
        {
            //System.out.println("ClipPlayer: usage:");
            //System.out.println("\tjava ClipPlayer <soundfile> <#loops>");
            File    clipFile = new File("../Pierce.wav");
            int     nLoopCount = 1;
            ClipPlayer  clipPlayer = new ClipPlayer(clipFile, nLoopCount);
        }
        else if (i == 1)
        {
            File    clipFile = new File(args[0]);
            int     nLoopCount = 1;
            ClipPlayer  clipPlayer = new ClipPlayer(clipFile, nLoopCount);
        }
        else
        {
            File    clipFile = new File(args[0]);
            int     nLoopCount = Integer.parseInt(args[1]);
            ClipPlayer  clipPlayer = new ClipPlayer(clipFile, nLoopCount);
        }
    }
}



/*** ClipPlayer.java ***/

